/*****************************************************************************
 * Implementation of a doubly-linked list management with embedded list head
 * structure, very similar to that in the Linux kernel
 *
 * Copyright (C) 2014 Parallels IP Holdings GmbH. All Rights Reserved.
 *
 * http://www.parallels.com
 *
 ****************************************************************************/
#ifndef __CD_LIST_H__
#define __CD_LIST_H__

#include <stddef.h>

#define cd_list_entry(ptr, type, member) ({			\
	const typeof(((type *)0)->member) *__xptr = (ptr);	\
	(type *)((char *)__xptr - offsetof(type, member));})

struct cd_list
{
	struct cd_list *next, *prev;
};

#define CD_LIST_INIT(name) { &(name), &(name) }
#define CD_LIST_HEAD(name) \
	struct cd_list name = CD_LIST_INIT(name)

#define cd_list_first_entry(ptr, type, member) \
	cd_list_entry((ptr)->next, type, member)

#define cd_list_last_entry(ptr, type, member) \
	cd_list_entry((ptr)->prev, type, member)

static inline void cd_list_init(struct cd_list *ptr)
{
	ptr->next = ptr;
	ptr->prev = ptr;
}

static inline void __cd_list_add(struct cd_list *newl,
	struct cd_list *prev, struct cd_list *next)
{
	next->prev = prev->next = newl;
	newl->next = next;
	newl->prev = prev;
}

/* add element to list head */
static inline void cd_list_add(struct cd_list *newl, struct cd_list *head)
{
	__cd_list_add(newl, head, head->next);
}

/* add element to list tail */
static inline void cd_list_add_tail(struct cd_list *newl, struct cd_list *head)
{
	__cd_list_add(newl, head->prev, head);
}

static inline void __cd_list_del(struct cd_list *prev, struct cd_list *next)
{
	next->prev = prev;
	prev->next = next;
}

static inline void cd_list_del(struct cd_list *elem)
{
	__cd_list_del(elem->prev, elem->next);
	cd_list_init(elem);
}

static inline int cd_list_empty(const struct cd_list* head)
{
	return head->next == head;
}

static inline void cd_list_move(struct cd_list *elem, struct cd_list *head)
{
	__cd_list_del(elem->prev, elem->next);
	cd_list_add(elem, head);
}

static inline void cd_list_move_tail(struct cd_list *elem, struct cd_list *head)
{
	__cd_list_del(elem->prev, elem->next);
	cd_list_add_tail(elem, head);
}

#define cd_list_for_each_entry(typeof_pos, pos, head, member) \
	for (pos = cd_list_entry((head)->next, typeof_pos, member); \
		&pos->member != (head); \
		pos = cd_list_entry(pos->member.next, typeof_pos, member))

#define cd_list_for_each_entry_reverse(typeof_pos, pos, head, member) \
	for (pos = cd_list_entry((head)->prev, typeof_pos, member); \
		&pos->member != (head); \
		pos = cd_list_entry(pos->member.prev, typeof_pos, member))

#define cd_list_for_each_entry_safe(typeof_pos, pos, n, head, member) \
	for (pos = cd_list_entry((head)->next, typeof_pos, member), \
		n = cd_list_entry(pos->member.next, typeof_pos, member); \
		&pos->member != (head); \
		pos = n, n = cd_list_entry(n->member.next, typeof_pos, member))

#define cd_list_for_each_entry_from_safe(typeof_pos, pos, n, head, member) \
	for (n = cd_list_entry(pos->member.next, typeof_pos, member); \
		&pos->member != (head); \
		pos = n, n = cd_list_entry(n->member.next, typeof_pos, member))

#define cd_list_for_each_entry_reverse_safe(typeof_pos, pos, p, head, member) \
	for (pos = cd_list_entry((head)->prev, typeof_pos, member),	\
		p = cd_list_entry(pos->member.prev, typeof_pos, member); \
		&pos->member != (head); \
		pos = p, p = cd_list_entry(pos->member.prev, typeof_pos, member))

#endif /* __CD_LIST_H__ */
