/*
 * In-place converter from Parallels to QCOW2 format
 *
 * Copyright (c) 2014 Denis V. Lunev
 * Copyright (c) 2024 Andrey Drobyshev
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <unistd.h>
#include <getopt.h>
#include <endian.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <malloc.h>
#include <stdint.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include "qcow2.h"
#include "parallels.h"
#include "util.h"

#define DIV_ROUND_UP(a, b) (((a) + (b) - 1) / (b))
#define ROUND_UP(a, b) (((a) + (b) - 1) / (b) * (b))
#define ARRAY_SIZE(x) (sizeof(x) / sizeof((x)[0]))

#define DEFAULT_REFCOUNT_ORDER 4

struct convert_ctx {
    int         fd;
    const char  *filename;

    uint8_t     *mapped;
    size_t      mmap_size;
    uint32_t    *bat;
    uint32_t    bat_size;
    uint64_t    bat_mult;

    uint32_t    cl_size;
    uint64_t    disk_size;

    uint64_t    data_off;
    uint64_t    max_data_off;
};

struct metadata_ctx {
    int         fd;
    const char  *filename;
    uint8_t     *mapped;
    size_t      mmap_size;
};

bool prl_verbose_output = false;
struct convert_ctx img_ctx = {
    .fd = -1,
    .filename = NULL,
    .mapped = NULL,
    .mmap_size = 0,
    0
};
struct metadata_ctx meta_ctx = {
    .fd = -1,
    .filename = NULL,
    .mapped = NULL,
    .mmap_size = 0
};

static void usage(FILE *fp)
{
    fprintf(fp, "%s"
"Usage: prl2qcow [-m, --meta META_IMAGE] [-s, --shift SHIFT_SIZE] PRL_IMAGE\n"
"   or: prl2qcow -r, --restore META_IMAGE QCOW_IMAGE\n"
"\n"
"Convert PRL_IMAGE from Parallels format to QCOW2 in-place.\n"
"If specified, save metadata (header + BAT) into META_IMAGE.\n"
"If specified, also shift virtual addresses within image by SHIFT_SIZE.\n"
"\n"
"Or restore QCOW_IMAGE to original Parallels image using META_IMAGE.\n"
"\n"
"  -s, --shift      shift size, in bytes or with unit suffix (k, K, Mb, GiB)\n"
"  -m, --meta       filename to save Parallels image metadata to\n"
"  -r, --restore    filename to restore Parallels image metadata from\n"
"                   (conflicts with -s, -m)\n"
"  -v, --verbose    enable debug output\n"
"  -h, --help       display this help and exit\n",
    fp == stderr ? "\n" : "");
}

static uint64_t get_file_size(int fd)
{
    struct stat st;

    if (fstat(fd, &st) < 0)
        return 0;
    return st.st_size;
}

static uint64_t prl_check_sign(const ParallelsHeader *header)
{
    if (!memcmp(header->magic, "WithouFreSpacExt", sizeof(header->magic)))
        return SECTOR_SIZE * header->tracks;
    if (!memcmp(header->magic, "WithoutFreeSpace", sizeof(header->magic)))
        return SECTOR_SIZE;

    err_exit("Unrecognized magic in %s", img_ctx.filename);
}

static void open_image(const char *filename, bool restore_mode)
{
    ParallelsHeader header;

    img_ctx.filename = filename;

    dprintf("Opening image: %s\n", img_ctx.filename);
    if ((img_ctx.fd = open(img_ctx.filename, O_RDWR)) < 0)
        perror_exit("Failed to open %s", img_ctx.filename);

    img_ctx.max_data_off = get_file_size(img_ctx.fd);
    if (img_ctx.max_data_off == 0)
        perror_exit("Failed to obtain file size of %s", img_ctx.filename);

    if (restore_mode) {
        /* We only need fd in restore mode. Check magic just in case. */
        uint32_t magic;
        dprintf("Checking image QCOW2 signature\n");
        if (read(img_ctx.fd, &magic, sizeof(magic)) < 0)
            perror_exit("Cannot read QCOW2 magic");
        magic = be32toh(magic);
        if (magic != QCOW_MAGIC)
            err_exit("QCOW2 magic in %s doesn't match", img_ctx.filename);
        lseek(img_ctx.fd, 0, SEEK_SET);
        return;
    }

    dprintf("Parsing Parallels header\n");
    if (read(img_ctx.fd, &header, sizeof(header)) < sizeof(header))
        perror_exit("Cannot read image header");

    img_ctx.bat_mult = prl_check_sign(&header);
    if (img_ctx.bat_mult == SECTOR_SIZE)
        /* remove possible garbage for old images */
        header.nb_sectors &= 0xffffffff;

    img_ctx.cl_size = header.tracks * SECTOR_SIZE;
    img_ctx.data_off = header.data_off * SECTOR_SIZE;
    img_ctx.disk_size = header.nb_sectors * SECTOR_SIZE;

    img_ctx.bat_size = header.nb_sectors / header.tracks;
    img_ctx.mmap_size = sizeof(header) + img_ctx.bat_size * sizeof(uint32_t);
    img_ctx.mapped = mmap(NULL, img_ctx.mmap_size, PROT_READ, MAP_SHARED,
                          img_ctx.fd, 0);
    if (img_ctx.mapped == MAP_FAILED)
        perror_exit("Failed to mmap metadata of image %s", img_ctx.filename);
    img_ctx.bat = (uint32_t *)(img_ctx.mapped + sizeof(header));

    dprintf("Cluster size: %u bytes\n", img_ctx.cl_size);
    dprintf("BAT size: %u entries\n", img_ctx.bat_size);
    dprintf("Virtual disk size: %lu bytes\n", img_ctx.disk_size);
    dprintf("Host image size: %lu bytes\n", img_ctx.max_data_off);
    dprintf("\n");
}

static void validate_image()
{
    int i;

    if ((img_ctx.cl_size - 1) & img_ctx.cl_size)
        err_exit("Cluster size %u isn't a power of 2", img_ctx.cl_size);

    if (img_ctx.data_off == 0)
        err_exit("Data offset is 0 (old image format, not aligned to "
                 "block size");

    if (img_ctx.max_data_off % img_ctx.cl_size != 0)
        err_exit("Image file size (max data offset) (%lu) must be aligned to "
                 "cluster size (%u)", img_ctx.max_data_off, img_ctx.cl_size);

    if (img_ctx.data_off % img_ctx.cl_size != 0)
        err_exit("Data offset (%lu) must be aligned to cluster size (%u)",
                 img_ctx.data_off, img_ctx.cl_size);

    for (i = 0; i < img_ctx.bat_size; i++) {
        uint64_t off = img_ctx.bat[i] * img_ctx.bat_mult;
        if (off % img_ctx.cl_size)
            err_exit("Cluster #%d offset (%lu) is not aligned to cluster "
                     "size (%u)", i, off, img_ctx.cl_size);
        if (off & ~L2E_OFFSET_MASK)
            err_exit("Cluster #%d offset (%lu) doesn't match L2E "
                     "mask (0x%llx)", i, off, L2E_OFFSET_MASK);
    }
}

static int ln2(unsigned size)
{
    int i;
    for (i = 0; i < 32 ; i++) {
        if ((1u << i) != size)
            continue;
        return i;
    }
    return -1;    /* should not happen, size is 2^n */
}

static uint64_t write_l2(int fd, uint64_t *l2, int count)
{
    uint64_t off;
    int i;

    for (i = 0; i < count; i++)
        if (l2[i] != htobe64(QCOW_OFLAG_ZERO))
            goto has_data;
    dprintf("only has unallocated clusters.  Nothing to write\n");
    return 0;

has_data:
    off = lseek(fd, 0, SEEK_END);
    dprintf("writing table (%lu bytes) at offset %lu\n",
            count * sizeof(uint64_t), off);
    write(fd, l2, count * sizeof(uint64_t));
    return off;
}

static void update_ref(uint16_t **refs, int nr_entries_ref_block,
                       uint64_t cl_size, uint64_t off)
{
    int cl_idx = off / cl_size;
    int ref_table_idx = cl_idx / nr_entries_ref_block;
    int ref_block_idx = cl_idx % nr_entries_ref_block;

    if (refs[ref_table_idx] == NULL) {
        refs[ref_table_idx] = (uint16_t *)malloc(cl_size);
        memset(refs[ref_table_idx], 0, cl_size);
    }
    refs[ref_table_idx][ref_block_idx] = htobe16(1);
}

static int refs_cl_count(uint16_t **refs, int nr_entries_ref_table)
{
    int i, count;

    for (count = i = 0; i < nr_entries_ref_table; i++) {
        if (refs[i] == NULL)
            continue;
        count++;
    }
    return count + 1;
}

static uint64_t shift(QCowHeader *header, uint64_t shift_bytes)
{
    size_t nr_new_clusters;

    if (shift_bytes == 0)
        return 0;

    assert(shift_bytes % img_ctx.cl_size == 0);
    nr_new_clusters = shift_bytes / img_ctx.cl_size;

    dprintf("\nRequested shift size: %lu bytes, %lu clusters\n",
            shift_bytes, nr_new_clusters);

    header->size = htobe64(be64toh(header->size) + shift_bytes);
    img_ctx.disk_size += shift_bytes;
    img_ctx.bat_size += nr_new_clusters;

    return nr_new_clusters;
}

static int convert(uint64_t shift_bytes)
{
    QCowHeader h;
    uint64_t nr_clusters, nr_l2_entries, shift_clusters, cl;
    uint64_t i, l1_size, off;
    int refcount_bits, nr_entries_ref_block, nr_entries_ref_table;
    size_t ref_table_size_aligned;
    uint64_t *l1, *l2;
    uint16_t **refs;
    uint64_t *ref_table;

    memset(&h, 0, sizeof(h));

    h.magic = htobe32(QCOW_MAGIC);
    h.version = htobe32(3);
    h.cluster_bits = htobe32(ln2(img_ctx.cl_size));
    h.size = htobe64(img_ctx.disk_size);

    /* The following fields are only valid for version >= 3 */
    /* these bits are dumb at the moment */
    h.incompatible_features = 0;
    h.compatible_features = 0;
    h.autoclear_features = 0;

    h.refcount_order = htobe32(DEFAULT_REFCOUNT_ORDER);
    h.header_length = htobe32(sizeof(h));

    shift_clusters = shift(&h, shift_bytes);
    nr_clusters = DIV_ROUND_UP(img_ctx.disk_size, img_ctx.cl_size) +
                  shift_clusters;
    nr_l2_entries = img_ctx.cl_size / sizeof(uint64_t);
    l1_size = DIV_ROUND_UP(nr_clusters, nr_l2_entries);
    h.l1_size = htobe32(l1_size);
    l1_size = ROUND_UP(sizeof(uint64_t) * l1_size, img_ctx.cl_size);

    l1 = (uint64_t *)malloc(l1_size);
    memset(l1, 0, l1_size);
    l2 = (uint64_t *)malloc(img_ctx.cl_size);

    refcount_bits = 1 << DEFAULT_REFCOUNT_ORDER; /* 16 by default */
    nr_entries_ref_block = img_ctx.cl_size * 8 / refcount_bits;
    nr_entries_ref_table = DIV_ROUND_UP(nr_clusters * 2, img_ctx.cl_size);
    refs = (uint16_t **)malloc(nr_entries_ref_table * sizeof(uint16_t *));
    memset(refs, 0, nr_entries_ref_table * sizeof(uint16_t *));

    /* Refcount table itself must be aligned to the cluster boundary. */
    ref_table_size_aligned = ROUND_UP(nr_entries_ref_table * sizeof(uint64_t),
                                      img_ctx.cl_size);
    ref_table = (uint64_t *)malloc(ref_table_size_aligned);
    memset(ref_table, 0, ref_table_size_aligned);

    dprintf("\nL1 table size: %u entries, %lu bytes (aligned to cluster)\n",
            be32toh(h.l1_size), l1_size);
    dprintf("L2 tables size: %lu entries, %u bytes (1 cluster)\n",
            nr_l2_entries, img_ctx.cl_size);
    dprintf("Refcount table size: %d entries, %lu bytes (aligned to "
            "cluster)\n", nr_entries_ref_table, ref_table_size_aligned);
    dprintf("Refcount blocks size: %d entries, %u bytes (1 cluster)\n",
            nr_entries_ref_block, img_ctx.cl_size);
    dprintf("\n");

    /* Update refcount for the very first host cluster containing header. */
    update_ref(refs, nr_entries_ref_block, img_ctx.cl_size, 0);

    /* create L2 tables */
    for (cl = 0; cl < img_ctx.bat_size; cl++) {
        uint64_t l2idx, l1idx, unshift_cl;

        l1idx = cl / nr_l2_entries;
        l2idx = cl % nr_l2_entries;

        if (l2idx == 0) {
            int i;
            if (l1idx != 0) {
                /* Dump filled L2 table and save L1 entry */
                dprintf("Filled L2 table #%lu: ", l1idx - 1);
                off = write_l2(img_ctx.fd, l2, nr_l2_entries);
                l1[l1idx - 1] = htobe64(off);
                update_ref(refs, nr_entries_ref_block, img_ctx.cl_size, off);
            }
            /* Initialize new L2 table */
            for (i = 0; i < nr_l2_entries; i++)
                l2[i] = htobe64(QCOW_OFLAG_ZERO);
        }

        /* There're no BAT entries for the empty clusters we're inserting
         * in the beginning of the image.
         */
        if (cl < shift_clusters)
            continue;

        unshift_cl = cl - shift_clusters;
        off = img_ctx.bat[unshift_cl] * img_ctx.bat_mult;

        if (img_ctx.bat[unshift_cl] == 0)
            continue;

        l2[l2idx] = htobe64(off | QCOW_OFLAG_COPIED);
        update_ref(refs, nr_entries_ref_block, img_ctx.cl_size, off);
    }

    /* For correct munmap() argument */
    img_ctx.bat_size -= shift_clusters;

    /* Dump remaining L2 and L1 entries */
    i = DIV_ROUND_UP(cl, nr_l2_entries) - 1;
    dprintf("Filled L2 table #%lu: ", i);
    off = write_l2(img_ctx.fd, l2, nr_l2_entries);
    l1[i] = htobe64(off | QCOW_OFLAG_COPIED);
    update_ref(refs, nr_entries_ref_block, img_ctx.cl_size, off);

    /* Dump L1 table and update its reference */
    off = get_file_size(img_ctx.fd);
    h.l1_table_offset = htobe64(off);
    update_ref(refs, nr_entries_ref_block, img_ctx.cl_size, off);
    dprintf("\nFilled L1 table: writing table (%lu bytes) at offset %lu\n\n",
            l1_size, off);
    write(img_ctx.fd, l1, l1_size);

    free(l1);
    free(l2);

    /* Refcount blocks should also have refcount 1 (each of them occupies
     * exactly 1 cluster).  When taking already allocated refcount blocks into
     * account new refcount blocks might emerge, so we're looping until their
     * number matches.  We're going to dump refcount blocks right at the
     * current EOF, one by one, hence the offsets.
     */
    off = get_file_size(img_ctx.fd);
    cl = 0;
    while (cl != refs_cl_count(refs, nr_entries_ref_table)) {
        cl = refs_cl_count(refs, nr_entries_ref_table);
        for (i = 0; i < cl; i++)
            update_ref(refs, nr_entries_ref_block, img_ctx.cl_size,
                       off + (img_ctx.cl_size * i));
    }

    /* Write refcount blocks one by one.  Save their actual file offsets in
     * the refcount table.
     */
    for (i = 0; i < nr_entries_ref_table; i++) {
        if (refs[i] == NULL) {
            dprintf("Refcount block #%lu is unallocated.  Skip writing\n", i);
            continue;
        }
        ref_table[i] = htobe64(get_file_size(img_ctx.fd));
        /* This might be redundant.  Just in case. */
        update_ref(refs, nr_entries_ref_block, img_ctx.cl_size,
                   get_file_size(img_ctx.fd));
        dprintf("Writing refcount block #%lu (%u bytes) at offset %lu\n",
                i, img_ctx.cl_size, be64toh(ref_table[i]));
        write(img_ctx.fd, refs[i], img_ctx.cl_size);
        free(refs[i]);
    }

    /* Dump refcount table itself. */
    h.refcount_table_offset = htobe64(get_file_size(img_ctx.fd));
    assert(be64toh(h.refcount_table_offset) % img_ctx.cl_size == 0);
    h.refcount_table_clusters = htobe32(ref_table_size_aligned /
                                        img_ctx.cl_size);
    dprintf("Writing refcount table (%lu bytes) at offset %lu\n",
            ref_table_size_aligned, be64toh(h.refcount_table_offset));
    write(img_ctx.fd, ref_table, ref_table_size_aligned);

    free(refs);
    free(ref_table);

    lseek(img_ctx.fd, 0, SEEK_SET);
    dprintf("\nWriting header (%lu bytes) at offset 0\n", sizeof(h));
    write(img_ctx.fd, &h, sizeof(h));

    /* there are no header extensions */
    off = 0;
    dprintf("Writing header extension end mark (0x%lx, %lu bytes) at "
            "offset %lu\n", off, sizeof(off), lseek(img_ctx.fd, 0, SEEK_CUR));
    write(img_ctx.fd, &off, sizeof(off));

    return 0;
}

static void open_meta(const char *metafile, PrlMetaHeader *header,
                      bool restore_mode)
{
    int flags = restore_mode ? O_RDONLY : (O_RDWR | O_CREAT | O_TRUNC);
    uint64_t meta_size;

    meta_ctx.filename = metafile;

    dprintf("Opening metadata image: %s\n", meta_ctx.filename);
    if ((meta_ctx.fd = open(meta_ctx.filename, flags)) < 0)
        perror_exit("Failed to open %s", meta_ctx.filename);

    if (!restore_mode)
        return;

    dprintf("Parsing metadata image header\n");
    if (read(meta_ctx.fd, header, sizeof(PrlMetaHeader)) < 0)
        perror_exit("Cannot read metadata image header");

    if (header->magic != PRL_META_MAGIC)
        err_exit("Unrecognized magic in %s", meta_ctx.filename);
    if (header->meta_header_size != sizeof(PrlMetaHeader))
        err_exit("Meta header size mismatch: %u != %lu",
                 header->meta_header_size, sizeof(PrlMetaHeader));

    if ((meta_size = get_file_size(meta_ctx.fd)) == 0)
        perror_exit("Failed to obtain metadata image size");
    if (header->meta_size != meta_size)
        err_exit("Metadata size missmatch: %lu != %lu",
                 header->meta_size, meta_size);
}

static ssize_t copy_meta(int fd, void *mapped, size_t size)
{
    ssize_t count;

    count = 0;
    while (count < size) {
        const size_t opt_wrsize = MIN(size, 128 * (1 << 10));
        size_t wrsize;
        ssize_t written;

        wrsize = MIN(opt_wrsize, size - count);
        if ((written = write(fd, mapped + count, wrsize)) < 0)
            return -1;
        count += written;
    }

    return count;
}

static void save_metadata(const char *metafile)
{
    PrlMetaHeader header;

    assert(metafile);

    open_meta(metafile, &header, false);

    header.magic = PRL_META_MAGIC;
    header.orig_img_size = img_ctx.max_data_off;
    header.meta_size = sizeof(header) + img_ctx.mmap_size;
    header.meta_header_size = sizeof(header);

    dprintf("Saving image metadata (%lu bytes) to %s at offset 0\n",
            header.meta_size, metafile);
    if (write(meta_ctx.fd, &header, sizeof(header)) < 0 ||
            copy_meta(meta_ctx.fd, img_ctx.mapped, img_ctx.mmap_size) < 0)
        perror_exit("Failed to write to %s", metafile);

    assert(get_file_size(meta_ctx.fd) == header.meta_size);
}

static void restore_metadata(const char *metafile)
{
    PrlMetaHeader header;
    void *prl_header_start;
    uint64_t mdsize;

    assert(metafile);

    open_meta(metafile, &header, true);

    if (header.orig_img_size > img_ctx.max_data_off)
        err_exit("Image became smaller.  That's impossible");

    meta_ctx.mmap_size = header.meta_size;
    meta_ctx.mapped = mmap(NULL, meta_ctx.mmap_size, PROT_READ, MAP_SHARED,
                           meta_ctx.fd, 0);
    if (meta_ctx.mapped == MAP_FAILED)
        perror_exit("Failed to mmap metadata image %s", meta_ctx.filename);

    prl_header_start = meta_ctx.mapped + sizeof(PrlMetaHeader);
    prl_check_sign(prl_header_start);
    mdsize = meta_ctx.mmap_size - sizeof(PrlMetaHeader);

    dprintf("Restoring image metadata (%lu bytes) from %s to offset 0\n",
            mdsize, meta_ctx.filename);
    if (copy_meta(img_ctx.fd, prl_header_start, mdsize) < 0)
        perror_exit("Failed to write to %s", img_ctx.filename);

    dprintf("Truncating image %s to original size %lu\n",
            img_ctx.filename, header.orig_img_size);
    if (ftruncate(img_ctx.fd, header.orig_img_size) < 0)
        perror_exit("Couldn't truncate image to original size");
}

static uint64_t suffix2mult(const char *endptr)
{
    const char *suffix[][5] = {
        {"k", "K", "Kb", "KB", "KiB"},
        {"m", "M", "Mb", "MB", "MiB"},
        {"g", "G", "Gb", "GB", "GiB"},
    };
    const uint64_t mult[] = {1 << 10, 1 << 20, 1 << 30};
    int i, j;

    assert(endptr && *endptr);
    for (i = 0; i < ARRAY_SIZE(suffix); ++i) {
        for (j = 0; j < ARRAY_SIZE(suffix[0]); ++j) {
            if (!strcmp(endptr, suffix[i][j]))
                return mult[i];
        }
    }

    return 0;
}

static uint64_t str2ull(const char *str)
{
    char *endptr;
    uint64_t res;

    while (*str && isspace(*str))
        str++;

    if (!*str)
        err_exit("Numeric value can't be empty");

    if (*str == '-')
        err_exit("Negative value isn't allowed: %s", str);
    if (str[0] == '0' && isdigit(str[1]))
        err_exit("Leading zeroes aren't allowed (base is 10): %s", str);

    errno = 0;
    res = strtoull(str, &endptr, 10);

    if (errno)
        perror_exit("Couldn't convert numeric value: %s", str);
    if (res == 0 && str[0] != '0')
        err_exit("Couldn't convert numeric value: %s", str);

    if (*endptr) {
        /* Trying to parse unit suffix */
        uint64_t mult = suffix2mult(endptr);
        if (mult == 0)
            err_exit("Couldn't convert numeric value: %s", str);
        if (res >= UINT64_MAX / mult)
            err_exit("Value %s is too big (integer overflow)", str);
        res *= mult;
    }

    return res;
}

__attribute__ ((destructor))
static void close_image()
{
    if (img_ctx.fd != -1) {
        fsync(img_ctx.fd);
        close(img_ctx.fd);
    }
    if (img_ctx.mapped)
        munmap(img_ctx.mapped, img_ctx.mmap_size);
    if (meta_ctx.fd != -1) {
        fsync(meta_ctx.fd);
        close(meta_ctx.fd);
    }
    if (meta_ctx.mapped)
        munmap(meta_ctx.mapped, meta_ctx.mmap_size);
}

int main(int argc, char* const argv[])
{
    int i, idx;
    uint64_t shift_bytes = 0;
    char *metafile = NULL;
    bool savemeta = false, restore = false;
    static struct option long_options[] = {
        {"shift",   required_argument, 0,  's'},
        {"meta",    required_argument, 0,  'm'},
        {"restore", required_argument, 0,  'r'},
        {"verbose", no_argument,       0,  'v'},
        {"help",    no_argument,       0,  'h'},
        {0,         0,                 0,   0}
    };

    while ((i = getopt_long(argc, argv, (char *)"s:vh", long_options,
            &idx)) != -1) {
        switch (i) {
        case 's':
            shift_bytes = str2ull(optarg);
            if (!shift_bytes)
                err_exit("Providing shift size 0 is pointless.\nEither "
                         "omit it or provide something meaningful");
            break;
        case 'm':
            metafile = optarg;
            savemeta = true;
            break;
        case 'r':
            metafile = optarg;
            restore = true;
            break;
        case 'v':
            prl_verbose_output = true;
            break;
        case 'h':
            usage(stdout);
            return 0;
        default:
            usage(stderr);
            exit(EXIT_FAILURE);
        }
    }

    if (argc < 2 || optind >= argc)
        err_exit("Image filename is not specified");
    if (savemeta && restore)
        err_exit("Options --meta and --restore aren't compatible");
    if (restore && shift_bytes != 0)
        err_exit("Providing shift size makes no sense in restore mode");

    open_image(argv[optind], restore);

    if (restore) {
        restore_metadata(metafile);
        return 0;
    }

    if (shift_bytes % img_ctx.cl_size != 0)
        err_exit("Shift size (%lu) must be divisible by cluster size (%u)",
                 shift_bytes, img_ctx.cl_size);

    validate_image();

    if (savemeta)
        save_metadata(metafile);

    convert(shift_bytes);
    close_image();

    return 0;
}
