/*****************************************************************************
 * A tool to convert Parallels virtual hard drives with "padding" to ones
 * without it.
 *
 * Copyright (C) 2014 Parallels IP Holdings GmbH. All Rights Reserved.
 *
 * http://www.parallels.com
 *
 ****************************************************************************/

/*****************************************************************************
 * Windows XP and 2003 by default align partitions to cylinder boundary, which
 * is 63 sectors.  As a result, most typical Windows XP/2k3 VM has its hard
 * drive with a single partition starting in sector 63, making I/O not
 * page-aligned, which is detrimental for performance.  To compensate for that,
 * Parallels virtual hard drives may have an option, "padding", meaning that
 * every guest LBA is accessed as LBA+1 on the host side.
 *
 * However, such images are hard to convert to other formats which know nothing
 * about "padding" (e.g. qcow2).
 *
 * This tool clears the "padding" in such virtual hdds, effectively "moving" all
 * the data by one sector, and then adjusts the partition table and possibly
 * other relevant stuff (e.g. filesystem superblocks) to reflect the new layout.
 *
 * The original "padded" hdd is left untouched so that revert is trivial; all
 * the changes are made in a new virtual hdd referring to the data files of the
 * original one as snapshots.
 *
 * The approach is the following:
 *
 * for every snapshot in the original hdd, which contains the blocks that need
 * adjusting, a new snapshot with a new data file is created with those blocks
 * corrected.  This new snapshot takes over the identity (guid and children) of
 * the original, while the original (referring to the original data file) is
 * made the parent of the new one, with a new identity and no other direct
 * children.
 *
 ****************************************************************************/

#define _GNU_SOURCE
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <endian.h>
#include <stdbool.h>
#include <stdint.h>
#include <assert.h>
#include <libgen.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xmlsave.h>
#include <libxml/xpath.h>
#include <uuid.h>

#include "parallels.h"
#include "cdlist.h"

#define _err(fmt, args...) fprintf(stderr, fmt "\n", ##args)
#if 1
#define _dbg(fmt, args...) fprintf(stderr, fmt "\n", ##args)
#else
#define _dbg(fmt, args...)
#endif

#define _cleanup_(x) __attribute__((cleanup(x)))

struct cmd_param {
    const char *src_dsc;
    const char *dst_dsc;
};

struct partition {
    uint8_t type;
    uint64_t start;
    uint64_t size;
};

struct img_descr {
    struct cd_list list;
    struct img_descr *parent;
    uuid_t guid;
    char *fname;
    bool is_new;
    struct partition parts[4];
    bool is_fixed;
};

struct img_file {
    int fd;
    void *map;
    ParallelsHeader *hdr;
    uint32_t *bat;
    uint32_t block_size;
    uint32_t bat_unit;
    uint32_t bat_entries;
    uint32_t data_start;
    off_t size;
};

static void init_img_file(struct img_file *file)
{
    memset(file, 0, sizeof(*file));
    file->fd = -1;
    file->map = MAP_FAILED;
}

static void clean_img_file(struct img_file *file)
{
    if (file->map != MAP_FAILED)
        munmap(file->map, file->data_start);
    if (file->fd >= 0)
        close(file->fd);
}

struct disk_descr {
    int src_fd;
    int src_dirfd;
    int dst_fd;
    int dst_dirfd;
    char *src_prefix;
    const char *dst_fname;
    xmlDocPtr doc;
    xmlNodePtr size_node;
    xmlNodePtr padding_node;
    xmlNodePtr storage_node;
    xmlNodePtr snapshots_node;
    unsigned long long size;
    unsigned heads;
    unsigned cylinders;
    unsigned sec_per_track;
    unsigned blocksize;
    struct cd_list img_head;
    bool is_error;
};

static void init_disk_descr(struct disk_descr *dsc)
{
    memset(dsc, 0, sizeof(*dsc));
    dsc->src_fd = -1;
    dsc->src_dirfd = -1;
    dsc->dst_fd = -1;
    dsc->dst_dirfd = -1;
    cd_list_init(&dsc->img_head);
    dsc->is_error = true;
}

static void clean_disk_descr(struct disk_descr *dsc)
{
    xmlFreeDoc(dsc->doc);
    while (!cd_list_empty(&dsc->img_head)) {
        struct img_descr *img =
            cd_list_first_entry(&dsc->img_head, struct
                img_descr, list);
        cd_list_del(&img->list);
        if (img->is_new && dsc->is_error)
            unlinkat(dsc->dst_dirfd, img->fname, 0);
        free(img->fname);
        free(img);
    }
    if (dsc->src_fd >= 0)
        close(dsc->src_fd);
    if (dsc->src_dirfd >= 0)
        close(dsc->src_dirfd);
    if (dsc->dst_fd >= 0)
        close(dsc->dst_fd);
    if (dsc->dst_dirfd >= 0)
        close(dsc->dst_dirfd);
    free(dsc->src_prefix);
    if (dsc->dst_fname && dsc->is_error)
        unlink(dsc->dst_fname);
}

static void usage(FILE *fp, const char *cmd)
{
    fprintf(fp, "Usage: %s SRC_DSC DST_DSC\n"
        "\n"
        "create new virtual disk without padding out of a disk with\n"
        "padding, using the original as snapshots\n"
        "\n"
        "  SRC_DSC\t- source disk descriptor (with padding)\n"
        "  DST_DSC\t- resulting disk descriptor (without padding)\n",
           cmd);
}

static int parse_cmd_param(int argc, const char **argv, struct cmd_param *cmdp)
{
    if (argc != 3) {
        usage(stderr, argv[0]);
        return -EINVAL;
    }

    cmdp->src_dsc = argv[1];
    cmdp->dst_dsc = argv[2];
    return 0;
}

static char *src_prefix(int src_fd, int dst_fd)
{
    struct stat src_stat, dst_stat;
    static const char tmpl[] = "/proc/self/fd/%d";
    char procfd[sizeof(tmpl) + 11];
    int i, n;

    if (fstat(src_fd, &src_stat) || fstat(dst_fd, &dst_stat)) {
        _err("fstat failed: %m");
        return NULL;
    }

    if (src_stat.st_dev == dst_stat.st_dev &&
        src_stat.st_ino == dst_stat.st_ino)
        return strdup("");

    sprintf(procfd, tmpl, src_fd);
    for (i = PATH_MAX; ; i *= 2) {
        char *ret = malloc(i);
        if (!ret)
            return NULL;

        n = readlink(procfd, ret, i - 2);
        if (n > 0) {
            ret[n] = '/';
            ret[n + 1] = '\0';
            return ret;
        }
        free(ret);
        if (errno != ENAMETOOLONG) {
            _err("readlink failed: %m");
            return NULL;
        }
    }
    return NULL;
}

static int setup_files(const struct cmd_param *cmdp, struct disk_descr *dsc)
{
    const char *dname;

    dsc->src_fd = open(cmdp->src_dsc, O_RDONLY);
    if (dsc->src_fd < 0) {
        int e = -errno;
        _err("fail to open file \"%s\": %m", cmdp->src_dsc);
        return e;
    }

    dsc->dst_fd = open(cmdp->dst_dsc, O_WRONLY | O_CREAT | O_EXCL, 0644);
    if (dsc->dst_fd < 0) {
        int e = -errno;
        _err("fail to open file \"%s\": %m", cmdp->dst_dsc);
        return e;
    }
    dsc->dst_fname = cmdp->dst_dsc;

    dname = dirname(strdupa(cmdp->src_dsc));
    dsc->src_dirfd = open(dname, O_DIRECTORY | O_PATH, 0);
    if (dsc->src_dirfd < 0) {
        int e = -errno;
        _err("fail to open dir \"%s\": %m", dname);
        return e;
    }

    dname = dirname(strdupa(cmdp->dst_dsc));
    dsc->dst_dirfd = open(dname, O_DIRECTORY | O_PATH, 0);
    if (dsc->dst_dirfd < 0) {
        int e = -errno;
        _err("fail to open dir \"%s\": %m", dname);
        return e;
    }

    dsc->src_prefix = src_prefix(dsc->src_dirfd, dsc->dst_dirfd);
    if (!dsc->src_prefix)
        return -ENOMEM;

    return 0;
}

static xmlXPathObjectPtr get_nodeset(xmlNodePtr root, char *expr)
{
    xmlXPathContextPtr ctx;
    xmlXPathObjectPtr ret;

    ctx = xmlXPathNewContext(root->doc);
    if (!ctx) {
        _err("xmlXPathNewContext: NULL");
        return NULL;
    }
    if (root)
        xmlXPathSetContextNode(root, ctx);

    ret = xmlXPathEval((xmlChar *) expr, ctx);
    xmlXPathFreeContext(ctx);
    if (!ret) {
        _err("xmlXPathEvalExpression(%s): NULL", expr);
        return NULL;
    }
    if (xmlXPathNodeSetIsEmpty(ret->nodesetval)) {
        xmlXPathFreeObject(ret);
        _err("xmlXPathEvalExpression(%s): empty result", expr);
        return NULL;
    }
    return ret;
}

static xmlNodePtr get_unique_node(xmlNodePtr root, char *expr)
{
    xmlNodePtr ret = NULL;
    xmlXPathObjectPtr nodes;

    nodes = get_nodeset(root, expr);
    if (!nodes)
        return NULL;

    if (xmlXPathNodeSetGetLength(nodes->nodesetval) != 1)
        _err("%s is non-unique", expr);
    else
        ret = xmlXPathNodeSetItem(nodes->nodesetval, 0);
    xmlXPathFreeObject(nodes);
    return ret;
}

static char *get_node_txt(xmlNodePtr node)
{
    xmlNodePtr child;

    if (!node)
        return NULL;

    child = node->children;
    if (node->type == XML_ELEMENT_NODE && child && !child->next &&
        child->type == XML_TEXT_NODE)
        return (char *) child->content;

    _err("\"%s\" is malformed (line %ld)", node->name,
        xmlGetLineNo(node));
    return NULL;
}

static int get_node_val(xmlNodePtr node,
    int (*getter)(char *, void *), void *val)
{
    int ret;
    char *txt, *savep;
    static const char *blank = " \t\n\r";

    txt = get_node_txt(node);
    if (!txt)
        return -ENOENT;

    txt = strdupa(txt);
    txt = strtok_r(txt, blank, &savep);
    if (!txt || strtok_r(NULL, blank, &savep))
        goto out_malformed;

    ret = getter(txt, val);
    if (ret)
        goto out_malformed;
    return ret;

out_malformed:
    _err("\"%s\" is malformed (line %ld)", node->name,
        xmlGetLineNo(node));
    return -EINVAL;
}

static int get_xp_val(xmlNodePtr root, char *expr,
    int (*getter)(char *, void *), void *val)
{
    return get_node_val(get_unique_node(root, expr), getter, val);
}

static int get_u(char *txt, void *val)
{
    errno = 0;
    * (unsigned *) val = strtoul(txt, NULL, 0);
    if (errno) {
        _err("failed to convert \"%s\" to unsigned long: %m", txt);
        return -EINVAL;
    }
    return 0;
}

static int get_ull(char *txt, void *val)
{
    errno = 0;
    * (unsigned long long *) val = strtoull(txt, NULL, 0);
    if (errno) {
        _err("failed to convert \"%s\" to unsigned long: %m", txt);
        return -EINVAL;
    }
    return 0;
}

static int get_uuid(char *txt, void *val)
{
    int ret = -EINVAL;
    size_t len = strlen(txt);
    if (len < 2)
        goto out_malformed;
    txt[len - 1] = '\0';
    txt++;
    ret = uuid_parse(txt, val);
    if (!ret)
        return 0;
out_malformed:
    _err("malformed uuid \"%s\"", txt);
    return ret;
}

static struct img_descr *new_img()
{
    struct img_descr *ret;

    ret = calloc(1, sizeof(*ret));
    if (!ret)
        _err("failed to allocate memory");
    else
        cd_list_init(&ret->list);
    return ret;
}

static int read_img(xmlNodePtr node, struct img_descr *img)
{
    char *type = NULL;
    xmlNodePtr c;

    for (c = node->children; c; c = c->next) {
        if (c->type != XML_ELEMENT_NODE)
            continue;
        if (!xmlStrcmp(c->name, BAD_CAST "GUID")) {
            int ret;
            if (!uuid_is_null(img->guid))
                goto out_duplicate;
            ret = get_node_val(c, get_uuid, img->guid);
            if (ret)
                return ret;
        } else if (!xmlStrcmp(c->name, BAD_CAST "File")) {
            char *fname;
            if (img->fname)
                goto out_duplicate;
            fname = get_node_txt(c);
            if (!fname)
                return -ENOENT;
            img->fname = strdup(fname);
            if (!img->fname)
                return -errno;
        } else if (!xmlStrcmp(c->name, BAD_CAST "Type")) {
            if (type)
                goto out_duplicate;
            type = get_node_txt(c);
            if (!type)
                return -ENOENT;
            if (strcmp(type, "Compressed")) {
                _err("invalid type \"%s\" (line %ld)",
                    type, xmlGetLineNo(c));
                return -EINVAL;
            }
        } else {
            _err("unexpected item \"%s\" (line %ld)",
                c->name, xmlGetLineNo(c));
            return -EINVAL;
        }
    }

    if (uuid_is_null(img->guid) || !type || !img->fname) {
        _err("incomplete item \"%s\" (line %ld)",
            node->name, xmlGetLineNo(node));
        return -EINVAL;
    }

    return 0;
out_duplicate:
    _err("duplicate item \"%s\" (line %ld)", c->name, xmlGetLineNo(c));
    return -EEXIST;
}

static int new_read_img(xmlNodePtr node, struct cd_list *head)
{
    struct img_descr *img;

    img = new_img();
    if (!img)
        return -ENOMEM;

    cd_list_add_tail(&img->list, head);

    return read_img(node, img);
}

static int read_snap(xmlNodePtr node, uuid_t guid, uuid_t parent)
{
    int ret;
    xmlNodePtr c;
    void *gp;

    uuid_clear(guid);
    uuid_clear(parent);

    for (c = node->children; c; c = c->next) {
        if (c->type != XML_ELEMENT_NODE)
            continue;
        if (!xmlStrcmp(c->name, BAD_CAST "GUID"))
            gp = guid;
        else if (!xmlStrcmp(c->name, BAD_CAST "ParentGUID"))
            gp = parent;
        else {
            _err("unexpected item \"%s\" (line %ld)",
                c->name, xmlGetLineNo(c));
            return -EINVAL;
        }

        if (!uuid_is_null(gp)) {
            _err("duplicate item \"%s\" (line %ld)", c->name, xmlGetLineNo(c));
            return -EEXIST;
        }

        ret = get_node_val(c, get_uuid, gp);
        if (ret)
            return ret;
    }

    if (uuid_is_null(guid)) {
        _err("incomplete item \"%s\" (line %ld)",
            node->name, xmlGetLineNo(node));
        return -EINVAL;
    }

    return 0;
}

static int link_img(xmlNodePtr node, struct cd_list *head,
    struct cd_list **linked_tail)
{
    int ret;
    uuid_t guid, parent_guid;
    struct img_descr *img = NULL, *parent = NULL, *cur;
    bool seen = true;
    char uuidstr[37];

    ret = read_snap(node, guid, parent_guid);
    if (ret)
        return ret;
    uuid_unparse(guid, uuidstr);

    cd_list_for_each_entry(struct img_descr, cur, head, list) {
        if (cur->list.prev == *linked_tail)
            seen = false;
        if (!uuid_compare(cur->guid, guid)) {
            if (seen) {
                _err("duplicate snapshot \"%s\" (line %ld)",
                    uuidstr, xmlGetLineNo(node));
                return -EEXIST;
            }
            img = cur;
        } else if (!uuid_compare(cur->guid, parent_guid))
            parent = cur;
    }

    if (!img) {
        _err("image \"%s\" not found (line %ld)",
            uuidstr, xmlGetLineNo(node));
        return -ENOENT;
    }
    if (!uuid_is_null(parent_guid) && !parent) {
        _err("image \"%s\"'s parent not found (line %ld)",
            uuidstr, xmlGetLineNo(node));
        return -ENOENT;
    }

    img->parent = parent;
    cd_list_move(&img->list, *linked_tail);
    *linked_tail = &img->list;

    return 0;
}

static int fill_images(struct disk_descr *dsc)
{
    int ret;
    xmlNodePtr c, c_next;
    struct cd_list *linked_tail;

    for (c = dsc->storage_node->children; c; c = c_next) {
        c_next = c->next;
        if (c->type != XML_ELEMENT_NODE ||
            xmlStrcmp(c->name, BAD_CAST "Image"))
            continue;

        ret = new_read_img(c, &dsc->img_head);
        if (ret)
            return ret;

        xmlUnlinkNode(c);
        xmlFreeNode(c);
    }

    linked_tail = &dsc->img_head;
    for (c = dsc->snapshots_node->children; c; c = c_next) {
        c_next = c->next;
        if (c->type != XML_ELEMENT_NODE ||
            xmlStrcmp(c->name, BAD_CAST "Shot"))
            continue;

        ret = link_img(c, &dsc->img_head, &linked_tail);
        if (ret)
            return ret;

        xmlUnlinkNode(c);
        xmlFreeNode(c);
    }

    if (linked_tail->next != &dsc->img_head) {
        _err("some Images lack Shot entries");
        return -ENOENT;
    }

    return 0;
}

static void set_lba_geometry(struct disk_descr *dsc)
{
    /* BIOS always uses LBA mode and presents virtual geometry via int 0x13
     * regardless of the disk-reported geometry.  See
     * http://en.wikipedia.org/wiki/Logical_block_addressing#LBA-assisted_translation
     * */
    dsc->sec_per_track = 63;

    for (dsc->heads = 16; dsc->heads < 256; dsc->heads <<= 1)
        if (dsc->heads * dsc->sec_per_track * 1024 > dsc->size)
            goto found;
    dsc->heads = 255;
found:
    dsc->cylinders = dsc->size / (dsc->heads * dsc->sec_per_track);
}

static int read_disk_descr(struct disk_descr *dsc)
{
    unsigned long long sec_val;
    unsigned padding;
    xmlNodePtr root, parm;

#define _DO_OR_RET(expr) do {        \
        int ret = (expr);    \
        if (ret) return ret;    \
    } while (0)

    dsc->doc = xmlReadFd(dsc->src_fd, NULL, NULL, XML_PARSE_NOBLANKS);
    if (!dsc->doc) {
        _err("Failed to parse source disk descriptor");
        return -EINVAL;
    }
    root = xmlDocGetRootElement(dsc->doc);
    if (!root || xmlStrcmp(root->name, BAD_CAST "Parallels_disk_image")) {
        _err("Malformed source disk descriptor");
        return -EINVAL;
    }

    parm = get_unique_node(root, "./Disk_Parameters");
    if (!parm)
        return -ENOENT;

    dsc->size_node = get_unique_node(parm, "./Disk_size");
    dsc->padding_node = get_unique_node(parm, "./Padding");
    dsc->storage_node = get_unique_node(root, "./StorageData/Storage");
    dsc->snapshots_node = get_unique_node(root, "./Snapshots");
    if (!dsc->snapshots_node || !dsc->padding_node || !dsc->storage_node)
        return -ENOENT;

    _DO_OR_RET(get_node_val(dsc->padding_node, get_u, &padding));
    if (!padding) {
        _err("the image has no padding -- it ain't broken");
        return -EINVAL;
    }
    _DO_OR_RET(get_xp_val(dsc->storage_node, "./Start", get_ull,
            &sec_val));
    if (sec_val) {
        _err("the image doesn't start at zero");
        return -EINVAL;
    }
    _DO_OR_RET(get_xp_val(dsc->storage_node, "./End", get_ull,
            &dsc->size));
    _DO_OR_RET(get_node_val(dsc->size_node, get_u, &sec_val));
    if (sec_val >= dsc->size) {
        _err("the drive (%llu) doesn't fit in the image (%llu)",
            sec_val, dsc->size);
        return -EINVAL;
    }

    _DO_OR_RET(get_xp_val(dsc->storage_node, "./Blocksize", get_u,
            &dsc->blocksize));

    set_lba_geometry(dsc);
    _dbg("size: %llu CHS: %u, %u, %u blocksize: %u",
        dsc->size, dsc->cylinders, dsc->heads, dsc->sec_per_track,
        dsc->blocksize);

#undef _DO_OR_RET

    return fill_images(dsc);
}

static int open_img_file(struct img_file *file, const char *fname,
    struct disk_descr *dsc)
{
    uint32_t data_start;

    file->fd = openat(dsc->src_dirfd, fname, O_RDONLY);
    if (file->fd < 0) {
        int e = -errno;
        _err("fail to open file \"%s\": %m", fname);
        return e;
    }

    file->map = mmap(NULL, BAT_OFFSET, PROT_READ, MAP_SHARED,
        file->fd, 0);
    if (file->map == MAP_FAILED) {
        int e = -errno;
        _err("mmap(\"%s\", %lu): %m", fname, BAT_OFFSET);
        return e;
    }
    file->data_start = BAT_OFFSET;

    file->hdr = file->map;
    file->block_size = le32toh(file->hdr->tracks) * SECTOR_SIZE;
    file->bat_entries = le32toh(file->hdr->bat_entries);
    data_start = le32toh(file->hdr->data_off) * SECTOR_SIZE;

    if (!memcmp(file->hdr->magic, "WithouFreSpacExt",
            sizeof(file->hdr->magic)))
        file->bat_unit = file->block_size;
    else if (!memcmp(file->hdr->magic, "WithoutFreeSpace",
            sizeof(file->hdr->magic)))
        file->bat_unit = SECTOR_SIZE;
    else {
        _err("unrecognized magic in \"%s\"", fname);
        return -EINVAL;
    }

    if (file->block_size != dsc->blocksize * SECTOR_SIZE ||
        (unsigned long long) file->block_size * file->bat_entries !=
            dsc->size * SECTOR_SIZE) {
        _err("image header in \"%s\" doesn't match descriptor", fname);
        return -EINVAL;
    }

    if (data_start < BAT_OFFSET + file->bat_entries * sizeof(uint32_t)) {
        _err("data overlaps with BAT in \"%s\"", fname);
        return -EINVAL;
    }

    file->map = mremap(file->map, file->data_start, data_start,
        MREMAP_MAYMOVE);
    if (file->map == MAP_FAILED) {
        int e = -errno;
        _err("mremap(%u -> %u): %m", file->data_start, data_start);
        return e;
    }
    file->data_start = data_start;
    file->hdr = file->map;
    file->bat = file->map + BAT_OFFSET;
    _dbg("image: %s, bs=%u bat_entries=%u bat_unit=%u size=%llu data_start=%u",
        file->hdr->magic, file->block_size,
        file->bat_entries, file->bat_unit,
        (unsigned long long) file->block_size * file->bat_entries / SECTOR_SIZE,
        file->data_start);
    return 0;
}

static int clone_img_file(struct img_file *file, struct img_descr *img,
    struct img_file *file_src, const char *fname_src, int dirfd)
{
    int i;
    char *fname_base;

    fname_base = basename(strdupa(fname_src));

    for (i = 0; ; i++) {
        if (asprintf(&img->fname, "%s.%0*x", fname_base,
            (int) sizeof(int) * 2, i) < 0) {
            img->fname = NULL;
            _err("failed to allocate memory");
            return -ENOMEM;
        }

        file->fd = openat(dirfd, img->fname,
            O_RDWR | O_CREAT | O_EXCL, 0644);
        if (file->fd >= 0)
            break;
        if (errno != EEXIST || i == INT_MAX - 1) {
            int e = -errno;
            _err("fail to open file \"%s\": %m", img->fname);
            return e;
        }
    }

    img->is_new = true;
    file->size = file->data_start = file_src->data_start;
    if (ftruncate(file->fd, file->size) < 0) {
        int e = -errno;
        _err("fail to resize \"%s\" to %lld: %m", img->fname,
            (long long) file->size);
        return e;
    }

    file->map = mmap(NULL, file->data_start, PROT_WRITE, MAP_SHARED,
        file->fd, 0);
    if (file->map == MAP_FAILED) {
        int e = -errno;
        _err("mmap(\"%s\", %u): %m", img->fname, file->data_start);
        return e;
    }

    memcpy(file->map, file_src->map, BAT_OFFSET);
    file->hdr = file->map;
    file->bat = file->map + BAT_OFFSET;
    file->block_size = file_src->block_size;
    file->bat_entries = file_src->bat_entries;
    file->bat_unit = file_src->bat_unit;
    return 0;
}

static int insert_src_img(struct img_file *file, struct img_descr *img,
    struct img_file *file_src, int dirfd)

{
    struct img_descr *img_src;

    img_src = new_img();
    if (!img_src) {
        _err("failed to allocate memory");
        return -ENOMEM;
    }

    /* insert img_src entry referring to the original image file *before*
     * the current entry now referring to the new image, to not interfere
     * with the ongoing list traversal */
    cd_list_add(&img_src->list, &img->list);
    img_src->parent = img->parent;
    img->parent = img_src;
    uuid_generate(img_src->guid);
    img_src->fname = img->fname;

    return clone_img_file(file, img, file_src, img_src->fname, dirfd);
}

static uint32_t off2blockno(struct img_file *file, off_t off)
{
    assert (off < (off_t) file->bat_entries * file->block_size);
    return off / file->block_size;
}

static int map_block(void **p, struct img_file *file, off_t off, void *src)
{
    uint32_t blockno;
    off_t foff;
    bool need_memcpy = false;

    blockno = off2blockno(file, off);

    if (!file->bat[blockno]) {
        if (!src)
            return -EFAULT;

        foff = file->size;
        if (ftruncate(file->fd, file->size + file->block_size) < 0) {
            int e = -errno;
            _err("ftruncate: %m");
            return e;
        }
        file->size += file->block_size;

        file->bat[blockno] = foff / file->bat_unit;

        need_memcpy = true;
    } else
        foff = file->bat[blockno] * file->bat_unit;

    *p = mmap(NULL, file->block_size, src ? PROT_WRITE : PROT_READ,
        MAP_SHARED, file->fd, foff);
    if (*p == MAP_FAILED) {
        int e = -errno;
        _err("mmap(@%lld): %m", (long long) foff);
        return e;
    }

    if (need_memcpy)
        memcpy(*p, src, file->block_size);
    return 0;
}

#define MBR_PART_START    0x1be
#define PART_ENTRY_SIZE    0x10

static void lba2chs(uint32_t lba, uint8_t *chs,
    unsigned heads, unsigned sec_per_track)
{
    uint16_t c = lba / (heads * sec_per_track);
    if (c >> 10) {
        chs[0] = 0xfe;
        chs[1] = chs[2] = 0xff;
        goto out;
    }
    chs[0] = (lba / sec_per_track) % heads;
    chs[1] = ((lba % sec_per_track + 1) & 0x3f) | ((c & 0x300) >> 2);
    chs[2] = c & 0xff;

out:
    _dbg("lba: %u chs: %hhx %hhx %hhx", lba, chs[0], chs[1], chs[2]);
}

static int fix_part_ent(uint8_t *pe, struct partition *part,
    unsigned heads, unsigned sec_per_track)
{
    int i;
    uint32_t lba_data[2];

    for (i = 0; i < PART_ENTRY_SIZE; i++)
        if (pe[i])
            goto non_empty;
    return 0;
non_empty:
    part->type = pe[4];

    memcpy(lba_data, pe + 8, sizeof(lba_data));
    part->start = le32toh(lba_data[0]) + 1;
    lba_data[0] = htole32(part->start);
    part->size = le32toh(lba_data[1]);
    memcpy(pe + 8, lba_data, sizeof(lba_data));

    lba2chs(part->start, pe + 1, heads, sec_per_track);
    lba2chs(part->start + part->size - 1, pe + 5, heads, sec_per_track);

    return 0;
}

static int fix_mbr(struct img_descr *img, struct img_file *file,
    struct img_file *file_new, struct disk_descr *dsc)
{
    int ret, i;
    void *src, *dst;

    ret = map_block(&src, file, 0, NULL);
    if (ret)
        return ret;
    ret = map_block(&dst, file_new, 0, src);
    if (ret)
        goto unmap_src;

    memcpy(dst, src + SECTOR_SIZE, SECTOR_SIZE);
    if (memcmp(dst + SECTOR_SIZE - 2, "\x55\xaa", 2)) {
        _err("no MBR magic");
        ret = -EINVAL;
        goto unmap;
    }

    for (i = 0; i < 4; i++) {
        ret = fix_part_ent(dst + MBR_PART_START + PART_ENTRY_SIZE * i,
            &img->parts[i], dsc->heads, dsc->sec_per_track);
        if (ret)
            goto unmap;
    }

unmap:
    munmap(dst, file->block_size);
unmap_src:
    munmap(src, file->block_size);
    return ret;
}

static int maybe_fix_mbr(struct img_descr *img, struct img_file *file,
    struct img_file *file_new, struct disk_descr *dsc)
{
    int ret;

    if (file->bat[0]) {
        ret = insert_src_img(file_new, img, file, dsc->dst_dirfd);
        if (ret)
            return ret;

        ret = fix_mbr(img, file, file_new, dsc);
        if (ret)
            return ret;
    } else {
        if (!img->parent) {
            char uuidstr[37];
            uuid_unparse(img->guid, uuidstr);
            _err("image \"%s\": no block #0 and no parent",
                uuidstr);
            return -EINVAL;
        }
        memcpy(img->parts, img->parent->parts, sizeof(img->parts));
    }

    return 0;
}

static int maybe_fix_part(struct img_descr *img, struct img_file *file,
    struct img_file *file_new, struct disk_descr *dsc, uint32_t start)
{
    int ret;
    void *src, *dst, *sb;
    off_t start_b = start * SECTOR_SIZE;

    /* partition start sector isn't in the original file - nothing to do */
    if (!file->bat[off2blockno(file, start_b)])
        return 0;

    /* if the new snapshot hasn't been inserted yet, do it now */
    if (file_new->fd < 0) {
        ret = insert_src_img(file_new, img, file, dsc->dst_dirfd);
        if (ret)
            return ret;
    }

    ret = map_block(&src, file, start_b, NULL);
    if (ret)
        return ret;
    ret = map_block(&dst, file_new, start_b, src);
    if (ret)
        goto unmap_src;

    sb = dst + start_b % file->block_size;
    if (memcmp(sb + SECTOR_SIZE - 2, "\x55\xaa", 2)) {
        _err("partition in sector %u has no bootrecord magic", start);
        ret = -EINVAL;
        goto unmap;
    }
    if (memcmp(sb + 3, "NTFS    ", 8) &&
        memcmp(sb + 0x52, "FAT32   ", 8)) {
        _err("partition in sector %u is neither NTFS nor FAT32",
            start);
        ret = -EINVAL;
        goto unmap;
    }

    /* update "hidden sectors"; this is vital for bootability */
    start = htole32(start);
    memcpy(sb + 0x1c, &start, 4);

unmap:
    munmap(dst, file->block_size);
unmap_src:
    munmap(src, file->block_size);
    return ret;
}

/*
 * this is the core part of the scheme:
 *
 * for every snapshot which contains the blocks which need adjusting, a new
 * snapshot is created with those blocks corrected.  The new snapshot takes over
 * the identity (guid and children) of the original, and the original is made
 * the parent of the new one, with new identity and no other direct children
 */
static int fix_img(struct img_descr *img, struct disk_descr *dsc)
{
    int ret, i;
    _cleanup_(clean_img_file) struct img_file file;
    _cleanup_(clean_img_file) struct img_file file_new;

    init_img_file(&file);
    init_img_file(&file_new);

    ret = open_img_file(&file, img->fname, dsc);
    if (ret)
        return ret;

    ret = maybe_fix_mbr(img, &file, &file_new, dsc);
    if (ret)
        return ret;

    for (i = 0; i < 4; i++) {
        if (img->parts[i].size == 0)
            continue;

        ret = maybe_fix_part(img, &file, &file_new, dsc,
            img->parts[i].start);
        if (ret)
            return ret;
    }

    img->is_fixed = true;
    return 0;
}

static int do_unpad(struct disk_descr *dsc)
{
    struct img_descr *img, *img2;

    cd_list_for_each_entry_safe(struct img_descr, img, img2,
        &dsc->img_head, list) {
        /* traverse the singly-linked ancestry list until we reach
         * already fixed one or null, temporarily reversing the list on
         * the way in order to be able to iterate back */
        struct img_descr *img_child = NULL, *img_parent;
        while (1) {
            img_parent = img->parent;
            img->parent = img_child;
            if (!img_parent || img_parent->is_fixed)
                break;
            img_child = img;
            img = img_parent;
        }

        /* fix_img() may insert a new image entry into the graph, but
         * its connectivity and the relationships will remain, so it's
         * ok to continue the iteration */
        for (; img; img = img_child) {
            int ret;
            img_child = img->parent;
            img->parent = img_parent;
            ret = fix_img(img, dsc);
            if (ret)
                return ret;
            img_parent = img;
        }
    }

    return 0;
}

static xmlNodePtr addGuid(xmlNodePtr parent, uuid_t val, char *lbl)
{
    char buf[39];
    uuid_unparse(val, buf + 1);
    buf[0] = '{';
    buf[37] = '}';
    return xmlNewChild(parent, NULL, BAD_CAST lbl, BAD_CAST buf);
}

static int write_disk_descr(struct disk_descr *dsc)
{
    struct img_descr *img;
    uuid_t null_uu  = {};
    char numbuf[21];
    xmlSaveCtxtPtr ctx;
    int ret = 0;

    sprintf(numbuf, "%llu", dsc->size);
    xmlNodeSetContent(dsc->size_node, BAD_CAST numbuf);
    xmlNodeSetContent(dsc->padding_node, BAD_CAST "0");

    cd_list_for_each_entry(struct img_descr, img, &dsc->img_head, list) {
        xmlNodePtr inode = xmlNewChild(dsc->storage_node, NULL,
            BAD_CAST "Image", NULL);
        addGuid(inode, img->guid, "GUID");
        xmlNewChild(inode, NULL, BAD_CAST "Type",
            BAD_CAST "Compressed");

        if (img->is_new || img->fname[0] == '/')
            xmlNewChild(inode, NULL, BAD_CAST "File",
                BAD_CAST img->fname);
        else {
            inode = xmlNewChild(inode, NULL, BAD_CAST "File",
                BAD_CAST dsc->src_prefix);
            xmlNodeAddContent(inode, BAD_CAST img->fname);
        }
    }

    cd_list_for_each_entry(struct img_descr, img, &dsc->img_head, list) {
        xmlNodePtr snode = xmlNewChild(dsc->snapshots_node, NULL,
            BAD_CAST "Shot", NULL);
        addGuid(snode, img->guid, "GUID");
        addGuid(snode, img->parent ? img->parent->guid : null_uu,
            "ParentGUID");
    }

    ctx = xmlSaveToFd(dsc->dst_fd, NULL, XML_SAVE_FORMAT);
    if (!ctx) {
        _err("failed to allocate memory");
        return -ENOMEM;
    }
    if (xmlSaveDoc(ctx, dsc->doc) < 0) {
        _err("error writing to destination disk descriptor");
        ret = -EIO;
    }
    if (xmlSaveClose(ctx) < 0)
        if (!ret) {
            _err("error writing to destination disk descriptor");
            ret = -EIO;
        }

    return ret;
}

int main(int argc, const char **argv)
{
    struct cmd_param cmdp;
    _cleanup_(clean_disk_descr) struct disk_descr dsc;

    init_disk_descr(&dsc);

    if (parse_cmd_param(argc, argv, &cmdp))
        return 1;

    if (setup_files(&cmdp, &dsc))
        return 1;

    if (read_disk_descr(&dsc))
        return 1;

    if (do_unpad(&dsc))
        return 1;

    if (write_disk_descr(&dsc))
        return 1;

    dsc.is_error = false;
    return 0;
}
